class MaveController

	attr_reader :pipe

	  PIPE_RD = 0;   PIPE_WR = 1
	SELECT_RD = 0; SELECT_WR = 1; SELECT_EX = 2

	def initialize(params)
		@configs = params[:CONFIGS]

		# 各モデル生成
		@models = {}
		@models[:__CNTL__]	= self
		@models[:STATUS]	= MaveStatus.new({:CONFIGS => @configs}); $status = @models[:STATUS] # デバッグ用
		@models[:PREVIEW]	= MavePreview.new({:CONFIGS => @configs})
		@models[:FOLDERS]	= MaveFolders.new({:CONFIGS => @configs})
		@models[:ACCOUNTS]	= MaveAccounts.new({:CONFIGS => @configs})

		# 各フォルダ初期化
		@models[:ACCOUNTS].each {|account|
			[:INBOX_FOLDER, :DRAFT_FOLDER, :OUTBOX_FOLDER, :SENT_FOLDER, :TRASH_FOLDER].each {|folder|
				@models[:FOLDERS].open_folder(account[folder])
			}
		}

		@exited_pids = []										# 終了した子プロセスの pid キュー
		@editfile_pid = {}										# pid に紐付く編集ファイル

		# シグナルハンドラ定義
		@pipe = IO.pipe
		Signal.trap(:CHLD) {|sig|								# 子プロセスが死んだ(エディタが終了した)ら
			begin
				while(Process.waitpid(-1, Process::WNOHANG | Process::WUNTRACED))
					if($?.exited?)
						@exited_pids << $?.pid
					else
						raise('unknown signal')
					end
				end
			rescue Errno::ECHILD
			end
			@pipe[PIPE_WR].putc(0x1B); @pipe[PIPE_WR].putc(?P)	# :mi_pickup_message にて、編集ファイルを回収
		}

		# 代表ビュー生成
		@views = MaveViews.new(@models)

		@actions = {											# コントローラ自ら実行できるアクション一覧
			:mk_global_fetch_mail_pop		=> method(:fetch_mail_pop),
			:mk_global_send_mail_smtp		=> method(:send_mail_smtp),
			:mk_global_toggle_what_charset	=> method(:toggle_what_charset),
			:mk_global_kill_mave			=> method(:kill_mave),

			:mk_global_previous_account		=> method(:previous_account),
			:mk_global_next_account			=> method(:next_account),

			:mk_global_new					=> method(:new_someone),
			:mk_global_renew				=> method(:renew_someone),
			:mk_global_reply				=> method(:reply_someone),
			:mk_global_reply_to_all			=> method(:reply_to_all_someone),
			:mk_global_forward				=> method(:forward_someone),
			:mk_global_resend				=> method(:resend_someone),
			:mk_global_edit					=> method(:edit_someone),

			:mi_pickup_message				=> method(:pickup_message),
		}
	end

	def action(command)
		(it = @actions[command]) ? it.call : nil
	end

	#-----------------------------------------------------------
	#
	#	メイン
	#
	def main
		begin
			loop {
				@views.clean
				@views.update									# 代表ビューに再表示を命令
				command = @views.ready(@views.active.name)		# アクティブビュー(キーマップ)を指定して、キー入力待ち
				@views.active.action(command) or				# アクティブビューにコマンドを処理させる
							@views.action(command) or			# 代表ビューにコマンドを処理させる
										action(command)			# ビューが実装してないコマンドは、コントローラ自ら処理する
			}
		rescue
			@views.close
			print $!.message + "\n"
			print $!.backtrace.join("\n")
		end
	end

	#-----------------------------------------------------------
	#
	#	メール受信(POP)
	#
	def fetch_mail_pop
		@pop_directory = MaveDirectory.new({:CONFIGS => @configs, :PATH => @configs[:POP_DIRECTORY]}) unless(@pop_directory)
		n_mail = 0; max_n_mail = 9999
		@models[:ACCOUNTS].each {|account|
			@models[:STATUS].log([_("Account '%1$s' connecting POP server '%2$s'..."), account.name, account.pop_server])
			begin
				account.pop {|popmail|
					@models[:STATUS].log([_(popmail)]) and next if(popmail.is_a?(String))
					max_n_mail = popmail and next if(popmail.is_a?(Integer))
					halfname = @pop_directory.create_mailfile {|fh|
						popmail.pop {|line|
							fh.write(line)
						}
					}
					fullname = @pop_directory.path + '/' + halfname
					mail = MaveMail.new({:CONFIGS => @configs, :FILE => File.new(fullname)})

					@models[:FOLDERS].each {|folder|
						if(folder.bind?(mail))
							folder.add_mail(mail)
							@pop_directory.delete(halfname)
						end
					}

					n_mail += 1
					@models[:STATUS].log([_('Popped (%1$d/%2$d) from [%3$s] [%4$s]'), n_mail, max_n_mail, mail.from.decode_mh, mail.subject.decode_mh])
				}
			rescue
				@models[:STATUS].log([_('Failed to pop mail. reason=[%s]'), $!.message])
			end
			@models[:STATUS].log([_('%1$s mail%2$s popped.'), (it = n_mail) == 0 ? 'no' : it.to_s, it == 1 ? '' : 's'])
		}
	end

	#-----------------------------------------------------------
	#
	#	メール送信(SMTP)
	#
	def send_mail_smtp
		#### 未実装
	end

	#-----------------------------------------------------------
	#
	#	既定のメールアカウントを変更
	#
	def previous_account
		account = @models[:ACCOUNTS].previous
		@models[:STATUS].log([_('Change current account to [%s].'), account[:NAME]])
	end
	def next_account
		account = @models[:ACCOUNTS].next
		@models[:STATUS].log([_('Change current account to [%s].'), account[:NAME]])
	end
 
	#-----------------------------------------------------------
	#
	#	各種編集作業
	#
	def new_someone												# 新規にメッセージ作成
		if(@views.current.is_a?(MaveMail))
			@models[:STATUS].log([_('Create new message.')])
			edit_message({:MODE => :NEW, :MAIL => @views.current})
		end
	end

	def renew_someone											# 宛先を流用して新規にメッセージ作成
		@models[:STATUS].log([_('Function not provided yet.')])
	end

	def reply_someone											# 返信メッセージを編集
		if(@views.current.is_a?(MaveMail))
			@models[:STATUS].log([_('Create reply message.')])
			edit_message({:MODE => :REPLY, :MAIL => @views.current})
		end
	end

	def reply_to_all_someone									# 全員への返信メッセージを編集
		@models[:STATUS].log([_('Function not provided yet.')])
	end

	def forward_someone											# 転送メッセージを編集
		@models[:STATUS].log([_('Function not provided yet.')])
	end

	def resend_someone											# 再送メッセージを編集
		@models[:STATUS].log([_('Function not provided yet.')])
	end

	def edit_someone											# 既存のメッセージを編集
		if(@views.current.is_a?(MaveMail))
			@models[:STATUS].log([_('Edit message.')])
			edit_message({:MODE => :EDIT, :MAIL => @views.current})
		end
	end

	#-----------------------------------------------------------
	#
	#	新規にメッセージファイルをエディタで編集させる
	#
	def edit_message(params)
		@pop_directory = MaveDirectory.new({:CONFIGS => @configs, :PATH => @configs[:POP_DIRECTORY]}) unless(@pop_directory)
		@force_kill = false
		begin
			params[:ACCOUNT] = @models[:ACCOUNTS].regular
			params[:HALFNAME] = @pop_directory.create_mailfile {|fh|
				MavePseudoMail.new({:CONFIGS => @configs}.update(params)).pseudo_each {|line|
					fh.write(line + "\n")
				}
			}
			params[:FORMHASH] = @pop_directory.md5(params[:HALFNAME])
			params[:SOURCEHASH] = params[:MAIL].md5 if(params[:MODE] == :EDIT)	# 編集開始時点の、編集元のハッシュ

			unless(@configs[:EDITOR_TYPE] == 'forkexec')		# 一旦 Curse を閉じて、エディタアプリを起動する場合
				@views.close
				system(@configs[:EDITOR] % (@pop_directory.path + '/' + params[:HALFNAME]))
				_pickup_message(params)
				@views.reopen
			else												# 別途ウィンドウでエディタアプリを起動する場合
				pid = fork {
					exec(@configs[:EDITOR] % (@pop_directory.path + '/' + params[:HALFNAME]))
				}
				@editfile_pid[pid.to_s] = params
			end
		rescue
			@models[:STATUS].log([_('Failed to editing mail. reason=[%s]'), $!.message])
		end
	end

	#-----------------------------------------------------------
	#
	#	エディタで編集済みのメッセージファイルを回収する
	#
	def pickup_message
		while(pid = @exited_pids.shift)
			_pickup_message(@editfile_pid[pid.to_s])
			@editfile_pid.delete(pid.to_s)
		end
	end
	def _pickup_message(editfile)
		@pop_directory = MaveDirectory.new({:CONFIGS => @configs, :PATH => @configs[:POP_DIRECTORY]}) unless(@pop_directory)
		unless(editfile[:FORMHASH] == @pop_directory.md5(editfile[:HALFNAME]))	# (編集前 == 現在)？
			mail = MavePseudoMail.new({:CONFIGS => @configs, :FILE => File.new(@pop_directory.path + '/' + editfile[:HALFNAME])})
			unless(editfile[:MODE] == :EDIT)
				folder = @models[:FOLDERS].open_folder(editfile[:ACCOUNT][:DRAFT_FOLDER])
				folder.add_mail(mail)
				@models[:STATUS].log([_('The message was stored in the [%s] folder.'), folder.name])
			else
				if(editfile[:SOURCEHASH] == editfile[:MAIL].md5)				# (編集前 == 現在)？
					editfile[:MAIL].folder.overwrite_mail(mail, editfile[:MAIL])
					@models[:STATUS].log([_('The message was overwrited.')])
				else
					editfile[:MAIL].folder.add_mail(mail)
					@models[:STATUS].log([_('The message was stored as another one (edit collision was detected).')])
				end
			end
		else
			@models[:STATUS].log([_('The message was discarded.')])
		end
		@pop_directory.delete(editfile[:HALFNAME])
	end

	#-----------------------------------------------------------
	#
	#	キャラクタセット情報の表示、切り替え(デバッグ用)
	#
	def toggle_what_charset
		state = MaveMail.toggle_what_charset
		@models[:STATUS].log(['what charset: %s', state.to_s])
	end

	#-----------------------------------------------------------
	#
	#	終了する
	#
	def kill_mave
		unless((it = @editfile_pid.size) < 1 or @force_kill)
			@models[:STATUS].log([_('%1$d editor%2$s alive.'), it, it == 1 ? ' is' : 's are'])
			@force_kill = true
		else
			raise 'Quit.'
		end
	end
end

__END__

